/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wFreestreamResidual.h"
#include "wFreestream.h"

#include "wElement.h"
#include "wCache.h"
#include "wGauss.h"

using namespace tbox;
using namespace dart;

/**
 * @brief Build the residual vector, on one freestream boundary element
 *
 * b = \int psi * v_inf * n dV
 */
Eigen::VectorXd FreestreamResidual::build(Element const &e, std::vector<double> const &phi, Freestream const &fs)
{
    // Get pre-computed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();

    // Normal velocity
    double vn = fs.f->eval(e, phi, 0).dot(e.getNormal());

    // Integral of mass flux (density times normal velocity) through freestream boundary
    Eigen::VectorXd b = Eigen::VectorXd::Zero(e.nodes.size());
    for (size_t k = 0; k < gauss.getN(); ++k)
        b += cache.getSf(k) * vn * gauss.getW(k) * e.getDetJ(k);
    return b;
}

/**
 * @brief Build the gradient of the residual vector with respect to the angle of attack, on one freestream boundary element
 *
 * b = \int psi * dv_inf * n dV
 */
Eigen::VectorXd FreestreamResidual::buildGradientAoa(tbox::Element const &e, std::vector<double> const &phi, Freestream const &fs)
{
    // Get pre-computed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();

    // Gradient of normal velocity
    double dvn = fs.f->evalGrad(e, phi, 0).dot(e.getNormal());

    // Build mass flux gradient
    Eigen::VectorXd b = Eigen::VectorXd::Zero(e.nodes.size());
    for (size_t k = 0; k < gauss.getN(); ++k)
        b += cache.getSf(k) * dvn * gauss.getW(k) * e.getDetJ(k);
    return b;
}
