/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WWAKEELEMENT_H
#define WWAKEELEMENT_H

#include "dart.h"
#include "wObject.h"
#include "wElement.h"
#include <vector>
#include <map>

namespace dart
{

/**
 * @brief Kutta element
 *
 * Made up of two surface elements on the wake and the connected volume
 * elements. Only Line2 (surface) and Tri3 (volume) elements are curently
 * implemented.
 * Since Tri3 or Tetra4 are used, the evalutaion of the volume integrals is
 * performed at dummy Gauss point #0. If volume elements with non-constant shape
 * function derivatives are used, those derivates should be recomputed at the
 * true surface element Gauss points.
 * @authors Adrien Crovato
 */
class DART_API WakeElement : public fwk::wObject
{
public:
    size_t no;             ///< ID
    tbox::Element *surUpE; ///< Surface element "+" side
    tbox::Element *surLwE; ///< Surface element "-" side
    tbox::Element *volUpE; ///< Connected volume elements "+" side
    tbox::Element *volLwE; ///< Connected volume elements "-" side

    size_t nRow;   ///< number of rows in stiffness matrix
    size_t nColUp; ///< number of columns in in stiffness matrix "+" side
    size_t nColLw; ///< number of columns in in stiffness matrix "-" side
    size_t nDim;   ///< dimension of volume element

    std::vector<std::pair<size_t, tbox::Node *>> wkN; ///< map between local index and contributing wake global nodes
    std::map<size_t, size_t> vsMap;                   ///< map between surface node and volume node indices

    WakeElement(size_t _no, tbox::Element *&_surUpE, tbox::Element *&_surLwE, tbox::Element *&_volUpE, tbox::Element *&_volLwE, std::vector<std::pair<size_t, tbox::Node *>> &_wkN);

#ifndef SWIG
    virtual void write(std::ostream &out) const override;
#endif

private:
    void mapNodes();
};

} // namespace dart

#endif //WWAKEELEMENT_H
