#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Newton raphson method coupled with linear solver
# Amaury Bilocq

import numpy as np
from fwk.coloring import ccolors

def newton(f, x, maxIt, tol=1.0e-8):
    '''Newton procedure to solve the non linear set of  boundary layer equations.
    Boundary layer equations are parabolic in x -> downstream marching is used
    An implicit marching model is applied (Crank Nicolson) -> matrix inversion is performed
    '''
    # Numerical Jacobian
    def jaco(f, x):
        eps = 1.0e-8
        n = len(x)
        jac = np.zeros((n,n))
        xp = x.copy()
        xm = x.copy()
        for j in range(n):
            # perturb solution
            xTmp = x[j]
            xp[j] += eps
            xm[j] -= eps
            # compute jacobian
            jac[:,j] = (f(xp) - f(xm)) / (2 * eps)
            # reset
            xp[j] = xTmp
            xm[j] = xTmp
        return jac
    # Backtrack line search
    def ls(f, x, dx):
        a = 1.0
        ires = np.linalg.norm(f(x))
        while a > 1/64:
            res = np.linalg.norm(f(x + a * dx))
            if res < ires:
                break
            else:
                a /= 2
        return a
    # Solve
    for i in range(maxIt):
        # solve
        jac = jaco(f,x)
        dx = np.linalg.solve(jac, -f(x))
        # line search
        a = ls(f, x, dx)
        x += a * dx
        fx = f(x)
        err = np.linalg.norm(fx)
        # safeguard
        x[0] = max(x[0],0.)
        x[1] = max(x[1],1.0005)
        if i != 0 and err < tol:
            return x
    # Raise error but return solution and error
    raise RuntimeError(ccolors.ANSI_YELLOW + 'Newton solver - too many iterations!\n' + ccolors.ANSI_RESET, x, err)
